# ddt/_finite_difference.py
"""Finite-difference schemes for estimating snapshot time derivatives."""

__all__ = [
    "fwd1",
    "fwd2",
    "fwd3",
    "fwd4",
    "fwd5",
    "fwd6",
    "bwd1",
    "bwd2",
    "bwd3",
    "bwd4",
    "bwd5",
    "bwd6",
    "ctr2",
    "ctr4",
    "ctr6",
    "ord2",
    "ord4",
    "ord6",
    "ddt_uniform",
    "ddt_nonuniform",
    "ddt",
    "UniformFiniteDifferencer",
    "NonuniformFiniteDifferencer",
]

import types
import warnings
import numpy as np

from .. import errors
from ._base import DerivativeEstimatorTemplate


def _finite_difference(states, coeffs, mode, inputs=None):
    r"""Compute the first time derivative with a finite difference scheme.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
        It is assumed that the states are uniformly spaced in time, i.e.,
        :math:`t_{j+1} - t_j` is the same for all :math:`j`.
    coeffs : (order + 1,) ndarray
        Finite difference coefficients. The convergence order of the
        estimation is ``len(coeffs) - 1``; for example, four points are
        used to obtain a third-order time derivative estimate.
    mode : str
        Indicator for if this is a forward, backward, or central difference.
        **Options:**

        * ``'fwd'``: forward difference.
        * ``'bwd'``: backward difference.
        * ``'ctr'``: central difference.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - order) ndarray
        State snapshots, excluding the last ``len(coeffs) - 1`` snapshots.
    ddts : (r, k - order) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - order) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is provided.
    """
    r, k = states.shape
    order = len(coeffs) - 1

    if mode == "fwd":
        cols = slice(0, -order)
    elif mode == "bwd":
        cols = slice(order, None)
    elif mode == "ctr":
        margin = order // 2
        cols = slice(margin, -margin)
    else:
        raise ValueError(f"invalid finite difference mode '{mode}'")

    ddts = np.zeros((r, k - order))
    for j, coeff in enumerate(coeffs):
        if coeff != 0:
            ddts += coeff * states[:, j : k - order + j]

    if inputs is not None:
        return states[:, cols], ddts, inputs[..., cols]
    return states[:, cols], ddts


# Forward differences =========================================================
def fwd1(states: np.ndarray, dt: float, inputs=None):
    r"""First-order forward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{\delta t}(\q(t_{j+1}) - \q(t_j))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 1) ndarray
        State snapshots, excluding the last snapshot.
    ddts : (r, k - 1) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 1) or (k - 1,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-1, 1]) / dt
    return _finite_difference(states, coeffs, "fwd", inputs)


def fwd2(states: np.ndarray, dt: float, inputs=None):
    r"""Second-order forward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{2\delta t}(-3\q(t_j) + 4\q(t_{j+1}) - \q(t_{j+2}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 2) ndarray
        State snapshots, excluding the last two snapshots.
    ddts : (r, k - 2) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 2) or (k - 2,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-3, 4, -1]) / (2 * dt)
    return _finite_difference(states, coeffs, "fwd", inputs)


def fwd3(states: np.ndarray, dt: float, inputs=None):
    r"""Third-order forward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{6\delta t}
       (-11\q(t_j) + 18\q(t_{j+1}) - 9\q(t_{j+2}) + 2\q(t_{j+3}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 3) ndarray
        State snapshots, excluding the last three snapshots.
    ddts : (r, k - 3) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 3) or (k - 3,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-11, 18, -9, 2]) / (6 * dt)
    return _finite_difference(states, coeffs, "fwd", inputs)


def fwd4(states: np.ndarray, dt: float, inputs=None):
    r"""Fourth-order forward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{12\delta t}(-25\q(t_j) + 48\q(t_{j+1}) - 36\q(t_{j+2})
       + 16\q(t_{j+3}) - 3\q(t_{j+4}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 4) ndarray
        State snapshots, excluding the last four snapshots.
    ddts : (r, k - 4) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 4) or (k - 4,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-25, 48, -36, 16, -3]) / (12 * dt)
    return _finite_difference(states, coeffs, "fwd", inputs)


def fwd5(states: np.ndarray, dt: float, inputs=None):
    r"""Fifth-order forward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{60\delta t}(-137\q(t_j) + 300\q(t_{j+1})
       - 300\q(t_{j+2}) + 200\q(t_{j+3}) - 75\q(t_{j+4}) + 12\q(t_{j+5}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 5) ndarray
        State snapshots, excluding the last five snapshots.
    ddts : (r, k - 5) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 5) or (k - 5,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-137, 300, -300, 200, -75, 12]) / (60 * dt)
    return _finite_difference(states, coeffs, "fwd", inputs)


def fwd6(states: np.ndarray, dt: float, inputs=None):
    r"""Sixth-order forward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{60\delta t}(
       -147\q(t_j) + 360\q(t_{j+1}) - 450\q(t_{j+2}) + 400\q(t_{j+3})
       - 225\q(t_{j+4}) + 72\q(t_{j+5}) - 10\q(t_{j+6}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 6) ndarray
        State snapshots, excluding the last six snapshots.
    ddts : (r, k - 6) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 6) or (k - 6,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-147, 360, -450, 400, -225, 72, -10]) / (60 * dt)
    return _finite_difference(states, coeffs, "fwd", inputs)


# Backwards differences =======================================================
def bwd1(states: np.ndarray, dt: float, inputs=None):
    r"""First-order backward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{\delta t}(\q(t_j) - \q(t_{j-1}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 1) ndarray
        State snapshots, excluding the first snapshot.
    ddts : (r, k - 1) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 1) or (k - 1,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-1, 1]) / dt
    return _finite_difference(states, coeffs, "bwd", inputs)


def bwd2(states: np.ndarray, dt: float, inputs=None):
    r"""Second-order backward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{2\delta t}(3\q(t_j) - 4\q(t_{j-1}) + \q(t_{j-2}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 2) ndarray
        State snapshots, excluding the first two snapshots.
    ddts : (r, k - 2) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 2) or (k - 2,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([1, -4, 3]) / (2 * dt)
    return _finite_difference(states, coeffs, "bwd", inputs)


def bwd3(states: np.ndarray, dt: float, inputs=None):
    r"""Third-order backward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{6\delta t}
       (11\q(t_j) - 18\q(t_{j-1}) + 9\q(t_{j-2}) - 2\q(t_{j-3}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 3) ndarray
        State snapshots, excluding the first three snapshots.
    ddts : (r, k - 3) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 3) or (k - 3,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-2, 9, -18, 11]) / (6 * dt)
    return _finite_difference(states, coeffs, "bwd", inputs)


def bwd4(states: np.ndarray, dt: float, inputs=None):
    r"""Fourth-order backward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{12\delta t}(25\q(t_j) - 48\q(t_{j-1}) + 36\q(t_{j-2})
       - 16\q(t_{j-3}) + 3\q(t_{j-4}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 4) ndarray
        State snapshots, excluding the first four snapshots.
    ddts : (r, k - 4) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 4) or (k - 4,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([3, -16, 36, -48, 25]) / (12 * dt)
    return _finite_difference(states, coeffs, "bwd", inputs)


def bwd5(states: np.ndarray, dt: float, inputs=None):
    r"""Fifth-order backward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{60\delta t}(137\q(t_j) - 300\q(t_{j-1})
       + 300\q(t_{j-2}) - 200\q(t_{j-3}) + 75\q(t_{j-4}) - 12\q(t_{j-5}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 5) ndarray
        State snapshots, excluding the first five snapshots.
    ddts : (r, k - 5) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 5) or (k - 5,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-12, 75, -200, 300, -300, 137]) / (60 * dt)
    return _finite_difference(states, coeffs, "bwd", inputs)


def bwd6(states: np.ndarray, dt: float, inputs=None):
    r"""Sixth-order backward difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{60\delta t}(
       147\q(t_j) - 360\q(t_{j-1}) + 450\q(t_{j-2}) - 400\q(t_{j-3})
       + 225\q(t_{j-4}) - 72\q(t_{j-5}) + 10\q(t_{j-6})
       )

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 6) ndarray
        State snapshots, excluding the first six snapshots.
    ddts : (r, k - 6) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 6) or (k - 6,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([10, -72, 225, -400, 450, -360, 147]) / (60 * dt)
    return _finite_difference(states, coeffs, "bwd", inputs)


# Central differences ========================================================
def ctr2(states: np.ndarray, dt: float, inputs=None):
    r"""Second-order central difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{2\delta t}(\q(t_{j+1}) - \q(t_{j-1}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 2) ndarray
        State snapshots, excluding the first and last snapshots.
    ddts : (r, k - 2) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 2) or (k - 2,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-1, 0, 1]) / (2 * dt)
    return _finite_difference(states, coeffs, "ctr", inputs)


def ctr4(states: np.ndarray, dt: float, inputs=None):
    r"""Fourth-order central difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{12\delta t}
       (\q(t_{j-2}) - 8\q(t_{j-1}) + 8\q(t_{j+1}) - \q(t_{j+2}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 4) ndarray
        State snapshots, excluding the first two and last two snapshots.
    ddts : (r, k - 4) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 4) or (k - 4,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([1, -8, 0, 8, -1]) / (12 * dt)
    return _finite_difference(states, coeffs, "ctr", inputs)


def ctr6(states: np.ndarray, dt: float, inputs=None):
    r"""Sixth-order central difference for estimating the first derivative.

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       \approx \frac{1}{60\delta t}
       (-\q(t_{j-3}) + 9\q(t_{j-2}) - 45\q(t_{j-1})
       + 45\q(t_{j+1})) - 9\q(t_{j+2}) + \q(t_{j+3})

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    _states : (r, k - 6) ndarray
        State snapshots, excluding the first three and last three snapshots.
    ddts : (r, k - 6) ndarray
        Time derivative estimates corresponding to the state snapshots.
    _inputs : (m, k - 6) or (k - 6,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    coeffs = np.array([-1, 9, -45, 0, 45, -9, 1]) / (60 * dt)
    return _finite_difference(states, coeffs, "ctr", inputs)


# Mixed differences ===========================================================
def _mixed_differences(states, dt, order, inputs):
    """Mixed finite differences on a uniform time domain."""
    ddts = ddt_uniform(states, dt, order=order)
    if inputs is not None:
        return states, ddts, inputs
    return states, ddts


def ord2(states: np.ndarray, dt: float, inputs=None):
    r"""Second-order forward, central, and backward differences for estimating
    the first derivative.

    Central differences are used where possible; forward differences are used
    for the first point and backward differences are used for the last point:

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_0}
       &\approx \frac{1}{2\delta t}(-3\q(t_0) + 4\q(t_{1}) - \q(t_{2})),
       \\ \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       &\approx \frac{1}{2\delta t}(-\q(t_{j-1}) + \q(t_{j+1})),
       \quad j = 2,\ldots,k-2,
       \\ \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_{k-1}}
       &\approx \frac{1}{2\delta t}(\q(t_{k-3}) - 4\q(t_{k-2}) + 3\q(t_{k-1}))

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    states : (r, k) ndarray
        State snapshots.
    ddts : (r, k) ndarray
        Time derivative estimates corresponding to the state snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    return _mixed_differences(states, dt, 2, inputs)


def ord4(states: np.ndarray, dt: float, inputs=None):
    r"""Fourth-order forward, central, and backward differences for estimating
    the first derivative.

    Central differences are used where possible; forward differences are used
    for the first two points and backward differences are used for the last
    two points:

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_0}
       &\approx \frac{1}{12\delta t}(-25\q(t_0) + 48\q(t_{1}) - 36\q(t_{2})
       + 16\q(t_{3}) - 3\q(t_{4})),
       \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_1}
       &\approx \frac{1}{12\delta t}(-3\q(t_0) - 10\q(t_{1}) + 18\q(t_{2})
       - 6\q(t_{3}) + \q(t_{4})),
       \\ \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       &\approx \frac{1}{12\delta t}
       (\q(t_{j-2}) - 8\q(t_{j-1}) + 8\q(t_{j+1}) - \q(t_{j+2})),
       \quad j = 2,\ldots,k-3,
       \\ \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_{k-2}}
       &\approx \frac{1}{12\delta t}(-\q(t_{k-5}) + 6\q(t_{k-4})
       - 18\q(t_{k-3}) + 10\q(t_{k-2}) + 3\q(t_{k-1})),
       \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_{k-1}}
       &\approx \frac{1}{12\delta t}(3\q(t_{k-5}) - 16\q(t_{k-4})
       + 36\q(t_{k-3}) - 48\q(t_{k-2}) + 25\q(t_{k-1})),

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.


    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    states : (r, k) ndarray
        State snapshots.
    ddts : (r, k) ndarray
        Time derivative estimates corresponding to the state snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    return _mixed_differences(states, dt, 4, inputs)


def ord6(states: np.ndarray, dt: float, inputs=None):
    r"""Sixth-order forward, central, and backward differences for estimating
    the first derivative.

    Central differences are used where possible; forward differences are used
    for the first three points and backward differences are used for the last
    three points:

    .. math::
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_0}
       &\approx \frac{1}{60\delta t}(
       -147\q(t_0) + 360\q(t_{1}) - 450\q(t_{2}) + 400\q(t_{3})
       - 225\q(t_{4}) + 72\q(t_{5}) - 10\q(t_{6})),
       \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_1}
       &\approx \frac{1}{60\delta t}(
       -10\q(t_0) - 77\q(t_{1}) + 150\q(t_{2}) - 100\q(t_{3})
       + 50\q(t_{4}) - 15\q(t_{5}) + 2\q(t_{6})),
       \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_2}
       &\approx \frac{1}{60\delta t}(
       2\q(t_0) - 24\q(t_{1}) - 35\q(t_{2}) + 80\q(t_{3})
       - 30\q(t_{4}) + 8\q(t_{5}) - \q(t_{6})),
       \\ \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_j}
       &\approx \frac{1}{60\delta t}
       (-\q(t_{j-3}) + 9\q(t_{j-2}) - 45\q(t_{j-1})
       + 45\q(t_{j+1})) - 9\q(t_{j+2}) + \q(t_{j+3}),
       \quad j = 3, \ldots, k - 4,
       \\ \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_{k-3}}
       &\approx \frac{1}{60\delta t}(
       \q(t_{k-7}) - 8\q(t_{k-6}) + 30\q(t_{k-5}) - 80\q(t_{k-4})
       + 35\q(t_{k-3}) + 24\q(t_{k-2}) - 2\q(t_{k-1})),
       \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_{k-2}}
       &\approx \frac{1}{60\delta t}(
       -2\q(t_{k-7}) + 15\q(t_{k-6}) - 50\q(t_{k-5}) + 100\q(t_{k-4})
       - 150\q(t_{k-3}) + 77\q(t_{k-2}) + 10\q(t_{k-1})),
       \\
       \frac{\textup{d}}{\textup{d}t}\q(t)\bigg|_{t = t_{k-1}}
       &\approx \frac{1}{60\delta t}(
       10\q(t_{k-7}) - 72\q(t_{k-6}) + 225\q(t_{k-5}) - 400\q(t_{k-4})
       + 450\q(t_{k-3}) - 360\q(t_{k-2}) + 147\q(t_{k-1})),

    where :math:`\delta t = t_{j+1} - t_j` for all :math:`j`.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to the states, if applicable.

    Returns
    -------
    states : (r, k) ndarray
        State snapshots.
    ddts : (r, k) ndarray
        Time derivative estimates corresponding to the state snapshots.
    inputs : (m, k) or (k,) ndarray or None
        Inputs corresponding to ``_states``, if applicable.
        **Only returned** if ``inputs`` is not ``None``.
    """
    return _mixed_differences(states, dt, 6, inputs)


# Class interface ============================================================
class UniformFiniteDifferencer(DerivativeEstimatorTemplate):
    r"""Time derivative estimation with finite differences for state snapshots
    spaced uniformly in time.

    Parameters
    ----------
    time_domain : (k,) ndarray
        Time domain corresponding to the snapshot data.
        This class requires uniformly spaced time domains, see
        :class:`NonuniformFiniteDifferencer` for non-uniform domains.
    scheme : str or callable
        Finite difference scheme to use.
        **Options:**

        * ``'fwd1'``: first-order forward differences, see :func:`fwd1`.
        * ``'fwd2'``: second-order forward differences, see :func:`fwd2`.
        * ``'fwd3'``: third-order forward differences, see :func:`fwd3`.
        * ``'fwd4'``: fourth-order forward differences, see :func:`fwd4`.
        * ``'fwd5'``: fifth-order forward differences, see :func:`fwd5`.
        * ``'fwd6'``: sixth-order forward differences, see :func:`fwd6`.
        * ``'bwd1'``: first-order backward differences, see :func:`bwd1`.
        * ``'bwd2'``: second-order backward differences, see :func:`bwd2`.
        * ``'bwd3'``: third-order backward differences, see :func:`bwd3`.
        * ``'bwd4'``: fourth-order backward differences, see :func:`bwd4`.
        * ``'bwd5'``: fifth-order backward differences, see :func:`bwd5`.
        * ``'bwd6'``: sixth-order backward differences, see :func:`bwd6`.
        * ``'ctr2'``: second-order backward differences, see :func:`ctr2`.
        * ``'ctr4'``: fourth-order backward differences, see :func:`ctr4`.
        * ``'ctr6'``: sixth-order backward differences, see :func:`ctr6`.
        * ``'ord2'``: second-order differences, see :func:`ord2`.
        * ``'ord4'``: fourth-order differences, see :func:`ord4`.
        * ``'ord6'``: sixth-order differences, see :func:`ord6`.

        If ``scheme`` is a callable function, its signature must match the
        following syntax.

        .. code-block:: python

           _states, ddts = scheme(states, dt)
           _states, ddts, _inputs = scheme(states, dt, inputs)

        Here ``dt`` is a positive float, the uniform time step.
        Each output should have the same number of columns.
    """

    _schemes = types.MappingProxyType(
        {
            "fwd1": fwd1,
            "fwd2": fwd2,
            "fwd3": fwd3,
            "fwd4": fwd4,
            "fwd5": fwd5,
            "fwd6": fwd6,
            "bwd1": bwd1,
            "bwd2": bwd2,
            "bwd3": bwd3,
            "bwd4": bwd4,
            "bwd5": bwd5,
            "bwd6": bwd6,
            "ctr2": ctr2,
            "ctr4": ctr4,
            "ctr6": ctr6,
            "ord2": ord2,
            "ord4": ord4,
            "ord6": ord6,
        }
    )

    def __init__(self, time_domain, scheme="ord4"):
        """Store the time domain and set the finite difference scheme."""
        DerivativeEstimatorTemplate.__init__(self, time_domain)

        # Check for uniform spacing.
        diffs = np.diff(time_domain)
        if not np.allclose(diffs, diffs[0]):
            raise ValueError("time domain must be uniformly spaced")

        # Set the finite difference scheme.
        if not callable(scheme):
            if scheme not in self._schemes:
                raise ValueError(
                    f"invalid finite difference scheme '{scheme}'"
                )
            scheme = self._schemes[scheme]
        self.__scheme = scheme

    # Properties --------------------------------------------------------------
    @property
    def dt(self):
        """Time step."""
        t = self.time_domain
        return t[1] - t[0]

    @property
    def scheme(self):
        """Finite difference engine."""
        return self.__scheme

    def __str__(self):
        return "\n  ".join(
            [
                DerivativeEstimatorTemplate.__str__(self),
                f"dt:          {self.dt:.4e}",
                f"scheme:      {self.scheme.__name__}()",
            ]
        )

    # Main routine ------------------------------------------------------------
    def estimate(self, states, inputs=None):
        r"""Estimate the first time derivatives of the states using
        finite differences.

        The stencil and order of the method are determined by the ``scheme``
        attribute.

        Parameters
        ----------
        states : (r, k) ndarray
            State snapshots, either full or (preferably) reduced.
        inputs : (m, k) ndarray or None
            Inputs corresponding to the state snapshots, if applicable.

        Returns
        -------
        _states : (r, k') ndarray
            Subset of the state snapshots.
        ddts : (r, k') ndarray
            First time derivatives corresponding to ``_states``.
        _inputs : (m, k') ndarray or None
            Inputs corresponding to ``_states``, if applicable.
            **Only returned** if ``inputs`` is provided.
        """
        states, inputs = self._check_dimensions(states, inputs, False)
        return self.scheme(states, self.dt, inputs)


class NonuniformFiniteDifferencer(DerivativeEstimatorTemplate):
    """Time derivative estimation with finite differences for state snapshots
    that are **not** spaced uniformly in time.

    This class essentially wraps :func:`numpy.gradient()`, which uses
    second-order finite differences.

    Parameters
    ----------
    time_domain : (k,) ndarray
        Time domain corresponding to the snapshot data.
        This class is for time domains that are not uniformly spaced, see
        :class:`UniformFiniteDifferencer` for uniformly spaced time domains.
    """

    def __init__(self, time_domain):
        """Set the time domain."""
        DerivativeEstimatorTemplate.__init__(self, time_domain)

        # Warn if time_domain in not uniform.
        if np.allclose(diffs := np.diff(time_domain), diffs[0]):
            warnings.warn(
                "time_domain is uniformly spaced, consider using "
                "UniformFiniteDifferencer",
                errors.OpInfWarning,
            )

    def __str__(self):
        """String representation: class name, time domain."""
        return "\n  ".join(
            [
                DerivativeEstimatorTemplate.__str__(self),
                "scheme:      np.gradient(edge_order=2)",
            ]
        )

    # Main routine ------------------------------------------------------------
    def estimate(self, states, inputs=None):
        r"""Estimate the first time derivatives of the states using
        second-order finite differences.

        Parameters
        ----------
        states : (r, k) ndarray
            State snapshots, either full or (preferably) reduced.
        inputs : (m, k) ndarray or None
            Inputs corresponding to the state snapshots, if applicable.

        Returns
        -------
        _states : (r, k) ndarray
            State snapshots.
        ddts : (r, k) ndarray
            First time derivatives corresponding to ``_states``.
        _inputs : (m, k) ndarray or None
            Inputs corresponding to ``_states``, if applicable.
            **Only returned** if ``inputs`` is provided.
        """
        states, inputs = self._check_dimensions(states, inputs)

        # Do the computation.
        ddts = np.gradient(states, self.time_domain, edge_order=2, axis=-1)

        if inputs is not None:
            return states, ddts, inputs
        return states, ddts


# Old API =====================================================================
def ddt_uniform(states, dt, order=2):
    """Forward, central, and backward differences for estimating the first
    derivative.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots.
    order : int {2, 4, 6}
        The order of the derivative approximation.

    Returns
    -------
    ddts : (r, k) ndarray
        Time derivative estimates corresponding to the state snapshots.
    """
    # Check dimensions and input types.
    if states.ndim != 2:
        raise errors.DimensionalityError("states must be two-dimensional")
    if not np.isscalar(dt):
        raise TypeError("time step dt must be a scalar (e.g., float)")

    ddts = np.empty_like(states)

    if order == 2:
        # TODO: Does this next line do what the rest does?
        # return np.gradient(states, dt, edge_order=2, axis=1)
        coeffs0 = np.array([-3, 4, -1]) / (2 * dt)
        # Forward differences on the front.
        ddts[:, 0] = states[:, :3] @ coeffs0
        # Central differences on the interior.
        _, ddts[:, 1:-1] = ctr2(states, dt, None)
        # Backward differences on the end.
        ddts[:, -1] = states[:, -3:] @ -coeffs0[::-1]

    elif order == 4:
        coeffs0 = np.array([-25, 48, -36, 16, -3]) / (12 * dt)
        coeffs1 = np.array([-3, -10, 18, -6, 1]) / (12 * dt)
        # Forward differences on the front.
        ddts[:, 0] = states[:, :5] @ coeffs0
        ddts[:, 1] = states[:, :5] @ coeffs1
        # Central differences on interior.
        _, ddts[:, 2:-2] = ctr4(states, dt, None)
        # Backward differences on the end.
        ddts[:, -2] = states[:, -5:] @ -coeffs1[::-1]
        ddts[:, -1] = states[:, -5:] @ -coeffs0[::-1]

    elif order == 6:
        coeffs0 = np.array([-147, 360, -450, 400, -225, 72, -10]) / (60 * dt)
        coeffs1 = np.array([-10, -77, 150, -100, 50, -15, 2]) / (60 * dt)
        coeffs2 = np.array([2, -24, -35, 80, -30, 8, -1]) / (60 * dt)
        # Forward differences on the front.
        ddts[:, 0] = states[:, :7] @ coeffs0
        ddts[:, 1] = states[:, :7] @ coeffs1
        ddts[:, 2] = states[:, :7] @ coeffs2
        # Central differences on interior.
        _, ddts[:, 3:-3] = ctr6(states, dt, None)
        # Backward differences on the end.
        ddts[:, -3] = states[:, -7:] @ -coeffs2[::-1]
        ddts[:, -2] = states[:, -7:] @ -coeffs1[::-1]
        ddts[:, -1] = states[:, -7:] @ -coeffs0[::-1]

    else:
        raise NotImplementedError(
            f"invalid order '{order}'; " "valid options: {2, 4, 6}"
        )

    return ddts


def ddt_nonuniform(states, t):
    """Second-order finite differences for estimating the first derivative.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    t : (k,) ndarray
        Time domain corresponding to the state snapshots.

    Returns
    -------
    ddts : (r, k) ndarray
        Time derivative estimates corresponding to the state snapshots.
    """
    # Check dimensions.
    if states.ndim != 2:
        raise errors.DimensionalityError("states must be two-dimensional")
    if t.ndim != 1:
        raise errors.DimensionalityError("time t must be one-dimensional")
    if states.shape[-1] != t.shape[0]:
        raise errors.DimensionalityError("states not aligned with time t")

    # Compute the derivative with a second-order difference scheme.
    return np.gradient(states, t, edge_order=2, axis=-1)


def ddt(states, *args, **kwargs):
    """Finite differences for estimating the first derivative.

    This is a convenience function that calls either :func:`ddt_uniform()` or
    :func:`ddt_nonuniform()`, depending on the arguments.

    Parameters
    ----------
    states : (r, k) ndarray
        State snapshots: ``states[:, j]`` is the state at time :math:`t_j`.
    dt : float
        Time step between snapshots (in the case of a time domain with
        uniform spacing).
    order : int {2, 4, 6} (optional)
        The order of the derivative approximation.
    t : (k,) ndarray
       Time domain corresponding to the state snapshots.
       May or may not be uniformly spaced.

    Returns
    -------
    ddts : (r, k) ndarray
        Time derivative estimates corresponding to the state snapshots.
    """
    n_args = len(args)  # Number of other positional args.
    n_kwargs = len(kwargs)  # Number of keyword args.
    n_total = n_args + n_kwargs  # Total number of other args.

    if n_total == 0:
        raise TypeError("at least one other argument required (dt or t)")
    elif n_total == 1:  # There is only one other argument.
        if n_kwargs == 1:  # It is a keyword argument.
            arg_name = list(kwargs.keys())[0]
            if arg_name == "dt":
                func = ddt_uniform
            elif arg_name == "t":
                func = ddt_nonuniform
            elif arg_name == "order":
                raise TypeError(
                    "keyword argument 'order' requires float " "argument dt"
                )
            else:
                raise TypeError(
                    f"ddt() got unexpected keyword argument '{arg_name}'"
                )
        elif n_args == 1:  # It is a positional argument.
            arg = args[0]
            if isinstance(arg, float):  # arg = dt.
                func = ddt_uniform
            elif isinstance(arg, np.ndarray):  # arg = t.
                func = ddt_nonuniform
            else:
                raise TypeError(f"invalid argument type '{type(arg)}'")
    elif n_total == 2:  # There are two other arguments: dt, order.
        func = ddt_uniform
    else:
        raise TypeError(
            "ddt() takes 2 or 3 positional arguments "
            f"but {n_total+1} were given"
        )

    return func(states, *args, **kwargs)


# Helper functions ============================================================
def _fdcoeffs(s):  # pragma: no cover
    r"""Vandermonde solve for finite difference coefficients.

    Parameters
    ----------
    s : (p,) ndarray
        Finite difference stencil. For example, ``s=[-2, -1, 0, 1]`` means
        we want coefficients :math:`c_{-2},c_{-1},c_{0},c_{1}` such that
        :math:`f'(x) \approx
        c_{-2} f(x - 2) + c_{-1} f(x - 1) + c_{0} f(x) + c_{1} f(x + 1)`.

    Returns
    -------
    (p,) ndarray
        Finite difference coefficients (the :math:`c_{j}`'s).
    """
    V = np.vander(s, increasing=True).T
    e = np.zeros_like(s)
    e[1] = 1
    return np.linalg.solve(V, e)


def _fdcoeffs2(s):  # pragma: no cover
    r"""Stable solve for finite difference coefficients (LeVeque).

    Parameters
    ----------
    s : (p,) ndarray
        Finite difference stencil. For example, ``[-2, -1, 0, 1]`` means
        get coefficients :math:`c_{-2},c_{-1},c_{0},c_{1}` such that
        :math:`f'(x) \approx
        c_{-2} f(x - 2) + c_{-1} f(x - 1) + c_{0} f(x) + c_{1} f(x + 1)`.

    Returns
    -------
    (p,) ndarray
        Finite difference coefficients (the :math:`c_{j}`'s).
    """
    pass



# models/mono/_nonparametric.py
"""Nonparametric monolithic dynamical systems models."""

__all__ = [
    # "SteadyModel",
    "DiscreteModel",
    "ContinuousModel",
]

import abc
import warnings
import numpy as np
import scipy.integrate as spintegrate
import scipy.interpolate as spinterpolate

from ._base import _Model
from ... import errors, utils, operators as _operators
from ...operators import _utils as oputils


# Base class ==================================================================
class _NonparametricModel(_Model):
    """Base class for nonparametric monolithic models.

    Parent class: :class:`opinf.models.mono._base._Model`

    Child classes:

    * :class:`opinf.models.DiscreteModel`
    * :class:`opinf.models.ContinuousModel`
    """

    _LHS_ARGNAME = "lhs"  # Name of LHS argument in fit(), e.g., "ddts".
    _LHS_LABEL = None  # String representation of LHS, e.g., "dq / dt".
    _STATE_LABEL = None  # String representation of state, e.g., "q(t)".
    _INPUT_LABEL = None  # String representation of input, e.g., "u(t)".

    # Properties: operators ---------------------------------------------------
    _operator_abbreviations = {
        "c": _operators.ConstantOperator,
        "A": _operators.LinearOperator,
        "H": _operators.QuadraticOperator,
        "G": _operators.CubicOperator,
        "B": _operators.InputOperator,
        "N": _operators.StateInputOperator,
    }

    @staticmethod
    def _isvalidoperator(op):
        """Return True if and only if ``op`` is a valid operator object
        for this class of model.
        """
        return oputils.is_nonparametric(op)

    @staticmethod
    def _check_operator_types_unique(ops):
        """Raise a ValueError if any two operators represent the same kind
        of operation (e.g., two constant operators).
        """
        if len({type(op) for op in ops}) != len(ops):
            raise ValueError("duplicate type in list of operators to infer")

    def _get_operator_of_type(self, OpClass):
        """Return the first operator of type ``OpClass``."""
        for op in self.operators:
            if isinstance(op, OpClass):
                return op

    # String representation ---------------------------------------------------
    def __str__(self):
        """String representation: structure of the model, dimensions, etc."""
        terms = [
            op._str(self._STATE_LABEL, self._INPUT_LABEL)
            for op in self.operators
        ]

        out = [
            self.__class__.__name__,
            f"structure: {self._LHS_LABEL} = " + " + ".join(terms),
            f"state_dimension: {self.state_dimension}",
            f"input_dimension: {self.input_dimension}",
            "operators:",
        ]
        for op in self.operators:
            out.append("  " + "\n    ".join(str(op).split("\n")))
        out.append("solver: " + "\n  ".join(str(self.solver).split("\n")))

        return "\n  ".join(out)

    def __repr__(self):
        """Unique ID + string representation."""
        return utils.str2repr(self)

    # Properties: operator inference ------------------------------------------
    @property
    def operator_matrix(self):
        r""":math:`r \times d(r, m)` operator matrix, e.g.,
        :math:`\Ohat = [~\chat~~\Ahat~~\Hhat~~\Bhat~]`.

        This matrix **does not** includes the entries of any operators whose
        entries are known *a priori*.
        """
        self._check_is_trained()
        return np.column_stack(
            [
                self.operators[i].entries
                for i in self._indices_of_operators_to_infer
            ]
        )

    # Fitting -----------------------------------------------------------------
    def _process_fit_arguments(self, states, lhs, inputs):
        """Prepare training data for Operator Inference by extracting
        dimensions, validating data sizes, and modifying the left-hand side
        data if there are any known operators.
        """
        # Clear non-intrusive operator data.
        self._clear()

        def _check_valid_dimension0(dataset, label):
            """Dimension 0 must be r (state dimensions)."""
            if (dim := dataset.shape[0]) != self.state_dimension:
                raise errors.DimensionalityError(
                    f"{label}.shape[0] = {dim} != r = {self.state_dimension}"
                )

        # Process states, extract model dimension if needed.
        states = np.atleast_2d(states)
        if self.state_dimension is None:
            self.state_dimension = states.shape[0]
        _check_valid_dimension0(states, "states")

        def _check_valid_dimension1(dataset, label):
            """Dimension 1 must be k (number of snapshots)."""
            if (dim := dataset.shape[1]) != (k := states.shape[1]):
                raise errors.DimensionalityError(
                    f"{label}.shape[-1] = {dim} != {k} = states.shape[-1]"
                )

        # Process LHS.
        lhs = np.atleast_2d(lhs)
        _check_valid_dimension0(lhs, self._LHS_ARGNAME)
        _check_valid_dimension1(lhs, self._LHS_ARGNAME)

        # Process inputs, extract input dimension if needed.
        self._check_inputargs(inputs, "inputs")
        if self._has_inputs:
            inputs = np.atleast_2d(inputs)
            if not self.input_dimension:
                self.input_dimension = inputs.shape[0]
            if inputs.shape[0] != self.input_dimension:
                raise errors.DimensionalityError(
                    f"inputs.shape[0] = {inputs.shape[0]} "
                    f"!= {self.input_dimension} = m"
                )
            _check_valid_dimension1(inputs, "inputs")

        # Subtract known operator evaluations from the LHS.
        for i in self._indices_of_known_operators:
            lhs = lhs - self.operators[i].apply(states, inputs)

        return states, lhs, inputs

    def _assemble_data_matrix(self, states, inputs):
        r"""Construct the Operator Inference data matrix :math:`\D`
        from state snapshots and/or input data.

        For example, if the model has the structure
        .. math::
            \ddt\qhat(t)
            = \chat + \Ahat\qhat(t)
            + \Hhat[\qhat(t)\otimes\qhat(t)] + \Bhat\u(t),

        then the data matrix is
        :math:`\D = [~
        \mathbf{1}~~
        \widehat{\Q}\trp~~
        (\widehat{\Q}\odot\widehat{\Q})\trp~~
        \U\trp~]`,

        where :math:`\widehat{\Q}` is ``states``
        and :math:`\U` is `inputs`.

        Parameters
        ----------
        states : (r, k) ndarray
            Snapshot training data.
        inputs : (m, k) ndarray or None
            Inputs corresponding to the snapshots.

        Returns
        -------
        D : (k, d(r, m)) ndarray
            Operator Inference data matrix.
        """
        return np.hstack(
            [
                self.operators[i].datablock(states, inputs).T
                for i in self._indices_of_operators_to_infer
            ]
        )

    def _extract_operators(self, Ohat):
        """Extract and save the inferred operators from the solution to the
        Operator Inference regression problem.

        Parameters
        ----------
        Ohat : (r, d(r, m)) ndarray
            Matrix of operator entries, concatenated horizontally.
        """
        index = 0
        for i in self._indices_of_operators_to_infer:
            endex = index + self.operators[i].operator_dimension(
                self.state_dimension, self.input_dimension
            )
            self.operators[i].set_entries(Ohat[:, index:endex])
            index = endex

    def _fit_solver(self, states, lhs, inputs=None):
        """Construct a solver object mapping the regularizer to solutions
        of the Operator Inference least-squares problem.
        """
        # Set up non-intrusive learning.
        states_, lhs_, inputs_ = self._process_fit_arguments(
            states, lhs, inputs
        )
        D = self._assemble_data_matrix(states_, inputs_)
        self.solver.fit(D, lhs_)

    def refit(self):
        """Solve the Operator Inference regression using the data from the
        last :meth:`fit()` call, then extract the inferred operators.

        This method is useful for calibrating the model operators with
        different ``solver`` hyperparameters without reprocessing any training
        data. For example, if ``solver`` is an :class:`opinf.lstsq.L2Solver`,
        changing its ``regularizer`` attribute and calling this method solves
        the regression with the new regression value without re-factorizing the
        data matrix.
        """
        # Fully intrusive case (nothing to learn).
        if self._fully_intrusive:
            warnings.warn(
                "all operators initialized explicitly, nothing to learn",
                errors.OpInfWarning,
            )
            return self

        # Execute non-intrusive learning.
        self._extract_operators(self.solver.solve())

    def fit(self, states, lhs, inputs=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ohat}\sum_{j=0}^{k-1}\left\|
           \fhat(\qhat_j, \u_j) - \zhat_j
           \right\|_2^2
           = \min_{\Ohat}\left\|\D\Ohat\trp - \dot{\Qhat}\trp\right\|_F^2

        where
        :math:`\zhat = \fhat(\qhat, \u)` is the model and

        * :math:`\qhat_j\in\RR^r` is a measurement of the state,
        * :math:`\u_j\in\RR^m` is a measurement of the input, and
        * :math:`\zhat_j\in\RR^r` is a measurement of the left-hand side
          of the model.

        The *operator matrix* :math:`\Ohat\in\RR^{r\times d(r,m)}` is such that
        :math:`\fhat(\q,\u) = \Ohat\d(\qhat,\u)` for some data vector
        :math:`\d(\qhat,\u)\in\RR^{d(r,m)}`; the *data matrix*
        :math:`\D\in\RR^{k\times d(r,m)}` is given by
        :math:`[~\d(\qhat_0,\u_0)~~\cdots~~\d(\qhat_{k-1},\u_{k-1})~]\trp`.
        Finally,
        :math:`\Zhat = [~\zhat_0~~\cdots~~\zhat_{k-1}~]\in\RR^{r\times k}`.
        See the :mod:`opinf.operators` module for more explanation.

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver`` set
        in the constructor.

        Parameters
        ----------
        states : (r, k) ndarray
            Snapshot training data. Each column is a single snapshot.
        lhs : (r, k) ndarray
            Left-hand side training data. Each column ``lhs[:, j]``
            corresponds to the snapshot ``states[:, j]``.
            The interpretation of this argument depends on the setting:
            forcing data for steady-state problems, next iteration for
            discrete-time problems, and time derivatives of the state for
            continuous-time problems.
        inputs : (m, k) or (k,) ndarray or None
            Input training data. Each column ``inputs[:, j]`` corresponds
            to the snapshot ``states[:, j]``.
            May be a one-dimensional array if ``m=1`` (scalar input).

        Returns
        -------
        self
        """
        # Fully intrusive case (nothing to learn).
        if self._fully_intrusive:
            warnings.warn(
                "all operators initialized explicitly, nothing to learn",
                errors.OpInfWarning,
            )
            return self

        self._fit_solver(states, lhs, inputs)
        self.refit()
        return self

    # Model evaluation --------------------------------------------------------
    def rhs(self, state, input_=None):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the function :math:`\Ophat(\qhat, \u)`
        where the model can be written as one of the following:

        * :math:`\ddt\qhat(t) = \Ophat(\qhat(t), \u(t))` (continuous time)
        * :math:`\qhat_{j+1} = \Ophat(\qhat_j, \u_j)` (discrete time)
        * :math:`\widehat{\mathbf{g}} = \Ophat(\qhat, \u)` (steady state)

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or None
            Input vector corresponding to the state.

        Returns
        -------
        out : (r,) ndarray
            Evaluation of the right-hand side of the model.
        """
        out = np.zeros_like(state)
        for op in self.operators:
            out += op.apply(state, input_)
        return out

    def jacobian(self, state, input_=None):
        r"""Construct and sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function :math:`\ddqhat\Ophat(\qhat, \u)`
        where the model can be written as one of the following:

        * :math:`\ddt\qhat(t) = \Ophat(\qhat(t), \u(t))` (continuous time)
        * :math:`\qhat_{j+1} = \Ophat(\qhat_{j}, \u_{j})` (discrete time)
        * :math:`\widehat{\mathbf{g}} = \Ophat(\qhat, \u)` (steady state)

        Parameters
        ----------
        state : (r,) ndarray
            State vector :math:`\qhat`.
        input_ : (m,) ndarray or None
            Input vector :math:`\u`.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.
        """
        r = self.state_dimension
        out = np.zeros_like(state, shape=(r, r))
        for op in self.operators:
            out += op.jacobian(state, input_)
        return out

    @abc.abstractmethod
    def predict(*args, **kwargs):  # pragma: no cover
        """Solve the model under specified conditions."""
        raise NotImplementedError

    # Model persistence -------------------------------------------------------
    def save(self, savefile, overwrite=False):
        """Serialize the model, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        with utils.hdf5_savehandle(savefile, overwrite=overwrite) as hf:
            # Metadata.
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["num_operators"] = len(self.operators)
            meta.attrs["r"] = (
                int(self.state_dimension) if self.state_dimension else 0
            )
            meta.attrs["m"] = (
                int(self.input_dimension) if self.input_dimension else 0
            )

            # Store operator data.
            hf.create_dataset(
                "indices_infer", data=self._indices_of_operators_to_infer
            )
            hf.create_dataset(
                "indices_known", data=self._indices_of_known_operators
            )
            for i, op in enumerate(self.operators):
                op.save(hf.create_group(f"operator_{i}"))
            if self.solver is not None:
                self.solver.save(hf.create_group("solver"))

    @classmethod
    def load(cls, loadfile: str):
        """Load a serialized model from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        model : _NonparametricModel
            Loaded model.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            # Load metadata.
            num_operators = int(hf["meta"].attrs["num_operators"])
            indices_infer = [int(i) for i in hf["indices_infer"][:]]
            indices_known = [int(i) for i in hf["indices_known"][:]]

            # Load operators.
            ops = []
            for i in range(num_operators):
                gp = hf[f"operator_{i}"]
                OpClass = getattr(_operators, gp["meta"].attrs["class"])
                ops.append(OpClass.load(gp))

            # Construct the model.
            model = cls(ops)
            model._indices_of_operators_to_infer = indices_infer
            model._indices_of_known_operators = indices_known
            if r := int(hf["meta"].attrs["r"]):
                model.state_dimension = r
            if (m := int(hf["meta"].attrs["m"])) and model._has_inputs:
                model.input_dimension = m

        return model


# Public classes ==============================================================
class SteadyModel(_NonparametricModel):  # pragma: no cover
    r"""Nonparametric steady state model :math:`\zhat = \fhat(\qhat)`.

    Here,

    * :math:`\qhat(t)\in\RR^{r}` is the model state, and
    * :math:`\u(t)\in\RR^{m}` is the (optional) input.

    The structure of :math:`\fhat` is specified through the
    ``operators`` argument.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
    solver : :mod:`opinf.lstsq` object or float > 0 or None
        Solver for the least-squares regression. Defaults:

        * ``None``: :class:`opinf.lstsq.PlainSolver`.
          SVD-based solve without regularization.
        * float > 0: :class:`opinf.lstsq.L2Solver`.
          SVD-based solve with scalar Tikhonov regularization.
    """

    _LHS_ARGNAME = "forcing"
    _LHS_LABEL = "z"
    _STATE_LABEL = "q"
    _INPUT_LABEL = None
    # TODO: disallow input terms?

    def fit(self, states, forcing=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ohat}\sum_{j=0}^{k-1}\left\|
           \fhat(\qhat_j) - \zhat_{j}
           \right\|_2^2
           = \min_{\Ohat}\left\|
           \D\Ohat\trp - [~\zhat_0~~\cdots~~\zhat_{k-1}~]\trp
           \right\|_F^2

        where :math:`\zhat = \fhat(\qhat)` is the model and

        * :math:`\qhat_j\in\RR^r` is a measurement of the state,
        * :math:`\zhat_j\in\RR^r` is a measurement of the forcing term
          corresponding to :math:`\qhat_j`,
        * :math:`\Ohat\in\RR^{r\times d(r,m)}` is the *operator matrix* such
          that :math:`\fhat(\q,\u) = \Ohat\d(\qhat,\u)` for some data vector
          :math:`\d(\qhat,\u)\in\RR^{d(r,m)}`, and
        * :math:`\D\in\RR^{k\times d(r,m)}` is the *data matrix* given by
          :math:`[~\d(\qhat_0,\u_0)~~\cdots~~\d(\qhat_{k-1},\u_{k-1})~]\trp`.

        See the :mod:`opinf.operators` module for further explanation.

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver``.

        Parameters
        ----------
        states : (r, k) ndarray
            Snapshot training data. Each column is a single snapshot.
        forcing : (r, k) ndarray or None
            Forcing training data. Each column ``forcing[:, j]``
            corresponds to the snapshot ``states[:, j]``.
            If ``None``, set ``forcing = 0``.

        Returns
        -------
        self
        """
        return _NonparametricModel.fit(self, states, forcing, inputs=None)

    def rhs(self, state):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the function :math:`\fhat(\qhat)` where the model is given by
        :math:`\zhat = \fhat(\qhat)`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector :math:`\qhat`.

        Returns
        -------
        g: (r,) ndarray
            Evaluation of the right-hand-side of the model.
        """
        return _NonparametricModel.rhs(self, state, None)

    def jacobian(self, state):
        r"""Sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function :math:`\ddqhat\fhat(\qhat)`
        where the model is given by :math:`\zhat = \fhat(\qhat)`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector :math:`\qhat`.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.
        """
        return _NonparametricModel.jacobian(self, state, input_=None)

    def predict(self, forcing, guess=None):
        """Solve the model with the given forcing and initial guess."""
        raise NotImplementedError("TODO")


class DiscreteModel(_NonparametricModel):
    r"""Nonparametric discrete dynamical system model
    :math:`\qhat_{j+1} = \fhat(\qhat_{j}, \u_{j})`.

    Here,

    * :math:`\qhat_j\in\RR^{r}` is the :math:`j`-th iteration
      of the model state, and
    * :math:`\u_j\in\RR^{m}` is the (optional) corresponding input.

    The structure of :math:`\fhat` is specified through the
    ``operators`` attribute.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
    solver : :mod:`opinf.lstsq` object or float > 0 or None
        Solver for the least-squares regression. Defaults:

        * ``None``: :class:`opinf.lstsq.PlainSolver`.
          SVD-based solve without regularization.
        * float > 0: :class:`opinf.lstsq.L2Solver`.
          SVD-based solve with scalar Tikhonov regularization.
    """

    _LHS_ARGNAME = "nextstates"
    _LHS_LABEL = r"q_{j+1}"
    _STATE_LABEL = r"q_{j}"
    _INPUT_LABEL = r"u_{j}"

    @staticmethod
    def stack_trajectories(statelist, inputlist=None):
        """Translate a collection of state trajectories and (optionally) inputs
        to arrays that are appropriate arguments for :meth:`fit()`.

        Parameters
        ----------
        statelist : list of s (r, k_i) ndarrays
            Collection of state trajectories.
        inputlist : list of s (m, k_i) ndarrays
            Collection of inputs corresponding to the state trajectories.

        Returns
        -------
        states : (r, sum_i(k_i)) ndarray
            Snapshot matrix with data from all but the final snapshot of each
            trajectory in ``statelist``.
        nextstates : (r, sum_i(k_i)) ndarray
            Snapshot matrix with data from all but the first snapshot of each
            trajectory in ``statelist``.
        inputs : (r, sum_i(k_i)) ndarray
            Input matrix with data from all but the last input for each
            trajectory. Only returned if ``inputlist`` is provided.
        """
        states = np.hstack([Q[:, :-1] for Q in statelist])
        nextstates = np.hstack([Q[:, 1:] for Q in statelist])
        if inputlist is not None:
            inputs = np.hstack(
                [
                    U[..., : (S.shape[1] - 1)]
                    for S, U in zip(statelist, inputlist)
                ]
            )
            return states, nextstates, inputs
        return states, nextstates

    def fit(self, states, nextstates=None, inputs=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ohat}\sum_{j=0}^{k-1}\left\|
           \fhat(\qhat_j, \u_j) - \zhat_{j}
           \right\|_2^2
           = \min_{\Ohat}\left\|
           \D\Ohat\trp - [~\zhat_0~~\cdots~~\zhat_{k-1}~]\trp
           \right\|_F^2

        where
        :math:`\zhat_j = \fhat(\qhat_j, \u_j)` is the model and

        * :math:`\qhat_j\in\RR^r` is a measurement of the state,
        * :math:`\u_j\in\RR^m` is a measurement of the input, and
        * :math:`\zhat_j\in\RR^r` is a measurement of the next state iteration
          after :math:`\qhat_j`,
        * :math:`\Ohat\in\RR^{r\times d(r,m)}` is the *operator matrix* such
          that :math:`\fhat(\q,\u) = \Ohat\d(\qhat,\u)` for some data vector
          :math:`\d(\qhat,\u)\in\RR^{d(r,m)}`, and
        * :math:`\D\in\RR^{k\times d(r,m)}` is the *data matrix* given by
          :math:`[~\d(\qhat_0,\u_0)~~\cdots~~\d(\qhat_{k-1},\u_{k-1})~]\trp`,

        See the :mod:`opinf.operators` module for further explanation.

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver``.

        Parameters
        ----------
        states : (r, k) ndarray
            Snapshot training data. Each column is a single snapshot.
        nextstates : (r, k) ndarray or None
            Next iteration training data. Each column ``nextstates[:, j]``
            is the iteration following ``states[:, j]``.
            If ``None``, use ``nextstates[:, j] = states[:, j+1]``.
        inputs : (m, k) or (k,) ndarray or None
            Input training data. Each column ``inputs[:, j]`` corresponds
            to the snapshot ``states[:, j]``.
            May be a one-dimensional array if ``m=1`` (scalar input).

        Returns
        -------
        self
        """
        if nextstates is None:
            nextstates = states[:, 1:]
            states = states[:, :-1]
        if inputs is not None:
            inputs = inputs[..., : states.shape[1]]
        return _NonparametricModel.fit(self, states, nextstates, inputs=inputs)

    def rhs(self, state, input_=None):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the function :math:`\fhat(\qhat, \u)`
        where the model is given by
        :math:`\qhat_{j+1} = \fhat(\qhat_{j}, \u_{j})`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector :math:`\qhat`.
        input_ : (m,) ndarray or None
            Input vector :math:`\u`.

        Returns
        -------
        nextstate : (r,) ndarray
            Evaluation of the right-hand side of the model.
        """
        return _NonparametricModel.rhs(self, state, input_)

    def jacobian(self, state, input_=None):
        r"""Sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function :math:`\ddqhat\fhat(\qhat, \u)`
        where the model is given by
        :math:`\qhat_{j+1} = \fhat(\qhat_{j}, \u_{j})`.

        Parameters
        ----------
        state : (r,) ndarray
            State vector :math:`\qhat`.
        input_ : (m,) ndarray or None
            Input vector :math:`\u`.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.
        """
        return _NonparametricModel.jacobian(self, state, input_)

    def predict(self, state0, niters, inputs=None):
        """Step forward the discrete dynamical system
        ``niters`` steps. Essentially, this amounts to the following.

        .. code-block:: python

           >>> states[:, 0] = state0
           >>> states[:, 1] = model.rhs(states[:, 0], inputs[:, 0])
           >>> states[:, 2] = model.rhs(states[:, 1], inputs[:, 1])
           ...                                     # Repeat `niters` times.

        Parameters
        ----------
        state0 : (r,) ndarray
            Initial state.
        niters : int
            Number of times to step the system forward.
        inputs : (m, niters-1) ndarray or None
            Inputs for the next ``niters - 1`` time steps.

        Returns
        -------
        states : (r, niters) ndarray
            Solution to the system, including the initial condition ``state0``.
        """
        self._check_is_trained()

        # Check initial condition dimension and process inputs.
        if (_shape := np.shape(state0)) != (self.state_dimension,):
            raise errors.DimensionalityError(
                "initial condition not aligned with model "
                f"(state0.shape = {_shape} != "
                f"({self.state_dimension},) = (r,))"
            )
        self._check_inputargs(inputs, "inputs")

        # Verify iteration argument.
        if not isinstance(niters, int) or niters < 1:
            raise ValueError("argument 'niters' must be a positive integer")

        # Create the solution array and fill in the initial condition.
        states = np.empty((self.state_dimension, niters))
        states[:, 0] = state0.copy()

        # Run the iteration.
        if self._has_inputs:
            if callable(inputs):
                raise TypeError("inputs must be NumPy array, not callable")

            # Validate shape of input, reshaping if input is 1d.
            U = np.atleast_2d(inputs)
            if (
                U.ndim != 2
                or U.shape[0] != self.input_dimension
                or U.shape[1] < niters - 1
            ):
                raise ValueError(
                    f"inputs.shape = ({U.shape} "
                    f"!= {(self.input_dimension, niters-1)} = (m, niters-1)"
                )
            for j in range(niters - 1):
                states[:, j + 1] = self.rhs(states[:, j], U[:, j])
        else:
            for j in range(niters - 1):
                states[:, j + 1] = self.rhs(states[:, j])

        # Return state results.
        return states


class ContinuousModel(_NonparametricModel):
    r"""Nonparametric system of ordinary differential equations
    :math:`\ddt\qhat(t) = \fhat(\qhat(t), \u(t))`.

    Here,

    * :math:`\qhat(t)\in\RR^{r}` is the model state, and
    * :math:`\u(t)\in\RR^{m}` is the (optional) input.

    The structure of :math:`\fhat` is specified through the
    ``operators`` argument.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
    solver : :mod:`opinf.lstsq` object or float > 0 or None
        Solver for the least-squares regression. Defaults:

        * ``None``: :class:`opinf.lstsq.PlainSolver`.
          SVD-based solve without regularization.
        * float > 0: :class:`opinf.lstsq.L2Solver`.
          SVD-based solve with scalar Tikhonov regularization.
    """

    _LHS_ARGNAME = "ddts"
    _LHS_LABEL = "dq / dt"
    _STATE_LABEL = "q(t)"
    _INPUT_LABEL = "u(t)"

    def fit(self, states, ddts, inputs=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ohat}\sum_{j=0}^{k-1}\left\|
           \fhat(\qhat_j, \u_j) - \dot{\qhat}_j
           \right\|_2^2
           = \min_{\Ohat}\left\|
           \D\Ohat\trp - [~\dot{\qhat}_0~~\cdots~~\dot{\qhat}_{k-1}~]\trp
           \right\|_F^2

        where
        :math:`\ddt\qhat(t) = \fhat(\qhat(t), \u(t))` is the model and

        * :math:`\qhat_j\in\RR^r` is the state at some time :math:`t_j`,
        * :math:`\u_j\in\RR^m` is the input at time :math:`t_j`,
        * :math:`\dot{\qhat}_j\in\RR^r` is the time derivative of the state
          at time :math:`t_j`, i.e.,
          :math:`\dot{\qhat}_j = \ddt\qhat(t)\big|_{t=t_j}`,
        * :math:`\Ohat\in\RR^{r\times d(r,m)}` is the *operator matrix* such
          that :math:`\fhat(\q,\u) = \Ohat\d(\qhat,\u)` for some data vector
          :math:`\d(\qhat,\u)\in\RR^{d(r,m)}`, and
        * :math:`\D\in\RR^{k\times d(r,m)}` is the *data matrix* given by
          :math:`[~\d(\qhat_0,\u_0)~~\cdots~~\d(\qhat_{k-1},\u_{k-1})~]\trp`.

        See the :mod:`opinf.operators` module for further explanation.

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver``.

        Parameters
        ----------
        states : (r, k) ndarray
            Snapshot training data. Each column is a single snapshot.
        ddts : (r, k) ndarray
            Snapshot time derivative data. Each column
            ``ddts[:, j]`` corresponds to the snapshot ``states[:, j]``.
        inputs : (m, k) or (k,) ndarray or None
            Input training data. Each column ``inputs[:, j]`` corresponds
            to the snapshot ``states[:, j]``.
            May be a one-dimensional array if ``m=1`` (scalar input).

        Returns
        -------
        self
        """
        return _NonparametricModel.fit(self, states, ddts, inputs=inputs)

    def rhs(self, t, state, input_func=None):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the right-hand side of the model, i.e., the function
        :math:`\fhat(\qhat(t), \u(t))` where the model is given by
        :math:`\ddt \qhat(t) = \fhat(\qhat(t), \u(t))`.

        Parameters
        ----------
        t : float
            Time :math:`t`, a scalar.
        state : (r,) ndarray
            State vector :math:`\qhat(t)` corresponding to time ``t``.
        input_func : callable(float) -> (m,), or None
            Input function that maps time ``t`` to the input vector
            :math:`\u(t)`.

        Returns
        -------
        dqdt : (r,) ndarray
            Evaluation of the right-hand side of the model.
        """
        input_ = None if not self._has_inputs else input_func(t)
        return _NonparametricModel.rhs(self, state, input_)

    def jacobian(self, t, state, input_func=None):
        r"""Sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function :math:`\ddqhat\fhat(\qhat(t), \u(t))`
        where the model is given by
        :math:`\ddt\qhat(t) = \fhat(\qhat(t), \u(t))`.

        Parameters
        ----------
        t : float
            Time :math:`t`, a scalar.
        state : (r,) ndarray
            State vector :math:`\qhat(t)` corresponding to time ``t``.
        input_func : callable(float) -> (m,), or None
            Input function that maps time ``t`` to the input vector
            :math:`\u(t)`.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.
        """
        input_ = None if not self._has_inputs else input_func(t)
        return _NonparametricModel.jacobian(self, state, input_)

    def predict(self, state0, t, input_func=None, **options):
        """Solve the system of ordinary differential equations.
        This method wraps :func:`scipy.integrate.solve_ivp()`.

        Parameters
        ----------
        state0 : (r,) ndarray
            Initial state vector.
        t : (nt,) ndarray
            Time domain over which to integrate the model.
        input_func : callable or (m, nt) ndarray
            Input as a function of time (preferred) or the input values at the
            times ``t``. If given as an array, cubic spline interpolation on
            the known data points is used as needed.
        options
            Arguments for :func:`scipy.integrate.solve_ivp()`.
            Common options:

            * **method : str** ODE solver for the model.

              * ``'RK45'`` (default): Explicit Runge--Kutta method
                of order 5(4).
              * ``'RK23'``: Explicit Runge--Kutta method
                of order 3(2).
              * ``'Radau'``: Implicit Runge--Kutta method
                of the Radau IIA family of order 5.
              * ``'BDF'``: Implicit multi-step variable-order (1 to 5) method
                based on a backward differentiation formula for the derivative.
              * ``'LSODA'``: Adams/BDF method with automatic stiffness
                detection and switching.

            * **max_step : float** Maximimum allowed integration step size.

        Returns
        -------
        states : (r, nt) ndarray
            Computed solution to the system over the time domain ``t``.
            A more detailed report on the integration results is stored as
            the ``predict_result_`` attribute.
        """
        self._check_is_trained()

        # Check initial condition dimension and process inputs.
        if (_shape := np.shape(state0)) != (self.state_dimension,):
            raise errors.DimensionalityError(
                "initial condition not aligned with model "
                f"(state0.shape = {_shape} != "
                f"({self.state_dimension},) = (r,))"
            )
        self._check_inputargs(input_func, "input_func")

        # Verify time domain.
        if t.ndim != 1:
            raise ValueError("time 't' must be one-dimensional")
        nt = t.shape[0]

        # Interpret control input argument.
        if self._has_inputs:
            if not callable(input_func):
                # input_func must be (m, nt) ndarray. Interpolate -> callable.
                U = np.atleast_2d(input_func)
                if U.shape != (self.input_dimension, nt):
                    raise ValueError(
                        f"input_func.shape = {U.shape} "
                        f"!= {(self.input_dimension, nt)} = (m, len(t))"
                    )
                input_func = spinterpolate.CubicSpline(t, U, axis=1)

            # Check dimension of input_func() outputs.
            _tmp = input_func(t[0])
            if self.input_dimension == 1 and np.isscalar(_tmp):
                original_input_func = input_func

                def input_func(t):
                    """Wrap outputs of input_func() as an array."""
                    return np.array([original_input_func(t)])

                _tmp = input_func(t[0])

            if not isinstance(_tmp, np.ndarray) or _tmp.shape != (
                self.input_dimension,
            ):
                raise errors.DimensionalityError(
                    "input_func() must return ndarray"
                    f" of shape (m,) = ({self.input_dimension},)"
                )

        if "method" in options and options["method"] in (
            # These methods require the Jacobian.
            "BDF",
            "Radau",
            "LSODA",
        ):
            options["jac"] = self.jacobian

        # Integrate the model.
        out = spintegrate.solve_ivp(
            self.rhs,  # Integrate this function
            [t[0], t[-1]],  # over this time interval
            state0,  # from this initial condition
            args=(input_func,),  # with this input function
            t_eval=t,  # evaluated at these points
            **options,  # using these solver options.
        )

        # Warn if the integration failed.
        if not out.success:  # pragma: no cover
            warnings.warn(out.message, spintegrate.IntegrationWarning)

        # Return state results.
        self.predict_result_ = out
        return out.y


# "Frozen" classes for parametric evaluation ==================================
class _FrozenMixin:
    """Mixin for evaluations of parametric models (disables fit())."""

    def _clear(self):
        raise NotImplementedError(
            "_clear() is disabled for this class, "
            "call fit() on the parametric model object"
        )

    @property
    def solver(self):
        return None

    @solver.setter
    def solver(self, solver):
        pass

    def fit(*args, **kwargs):
        raise NotImplementedError(
            "fit() is disabled for this class, "
            "call fit() on the parametric model object"
        )


class _FrozenSteadyModel(_FrozenMixin, SteadyModel):
    """Nonparametric steady-state model that is the evaluation of
    a parametric model.
    """

    pass


class _FrozenDiscreteModel(_FrozenMixin, DiscreteModel):
    """Nonparametric discrete-time model that is the evaluation of
    a parametric model.
    """

    pass


class _FrozenContinuousModel(_FrozenMixin, ContinuousModel):
    """Nonparametric continuous-time model that is the evaluation of
    a parametric model.
    """

    pass



# pre/_reprojection.py
"""Re-projection of trajectories for recovering intrusive models with
Operator Inference.
"""

__all__ = [
    "reproject_discrete",
    "reproject_continuous",
]

import numpy as np


# Reprojection schemes ========================================================
def reproject_discrete(f, basis, init, niters, inputs=None):
    """Sample re-projected trajectories of the discrete dynamical system

        q_{j+1} = f(q_{j}, u_{j}),  q_{0} = q0.

    Parameters
    ----------
    f : callable mapping (n,) ndarray (and (m,) ndarray) to (n,) ndarray
        Function defining the (full-order) discrete dynamical system. Accepts
        a full-order state vector and (optionally) an input vector and returns
        another full-order state vector.
    basis : (n, r) ndarray
        Basis for the low-dimensional linear subspace (e.g., POD basis).
    init : (n,) ndarray
        Initial condition for the iteration in the high-dimensional space.
    niters : int
        The number of iterations to do.
    inputs : (m, niters-1) or (niters-1) ndarray
        Control inputs, one for each iteration beyond the initial condition.

    Returns
    -------
    states_reprojected : (r, niters) ndarray
        Re-projected state trajectories in the projected low-dimensional space.
    """
    # Validate and extract dimensions.
    n, r = basis.shape
    if init.shape != (n,):
        raise ValueError("basis and initial condition not aligned")

    # Create the solution array and fill in the initial condition.
    states_ = np.empty((r, niters))
    states_[:, 0] = basis.T @ init

    # Run the re-projection iteration.
    if inputs is None:
        for j in range(niters - 1):
            states_[:, j + 1] = basis.T @ f(basis @ states_[:, j])
    elif inputs.ndim == 1:
        for j in range(niters - 1):
            states_[:, j + 1] = basis.T @ f(basis @ states_[:, j], inputs[j])
    else:
        for j in range(niters - 1):
            states_[:, j + 1] = basis.T @ f(
                basis @ states_[:, j], inputs[:, j]
            )

    return states_


def reproject_continuous(f, basis, states, inputs=None):
    """Sample re-projected trajectories of the continuous system of ODEs

        dq / dt = f(t, q(t), u(t)),     q(0) = q0.

    Parameters
    ----------
    f : callable mapping (n,) ndarray (and (m,) ndarray) to (n,) ndarray
        Function defining the (full-order) differential equation. Accepts a
        full-order state vector and (optionally) an input vector and returns
        another full-order state vector.
    basis : (n, r) ndarray
        Basis for the low-dimensional linear subspace.
    states : (n, k) ndarray
        State trajectories (training data).
    inputs : (m, k) or (k,) ndarray
        Control inputs corresponding to the state trajectories.

    Returns
    -------
    states_reprojected : (r, k) ndarray
        Re-projected state trajectories in the projected low-dimensional space.
    ddts_reprojected : (r, k) ndarray
        Re-projected velocities in the projected low-dimensional space.
    """
    # Validate and extract dimensions.
    if states.shape[0] != basis.shape[0]:
        raise ValueError(
            "states and basis not aligned, first dimension "
            f"{states.shape[0]} != {basis.shape[0]}"
        )
    n, r = basis.shape
    k = states.shape[1]

    # Create the solution arrays.
    states_ = basis.T @ states
    ddts_ = np.empty((r, k))

    # Run the re-projection iteration.
    if inputs is None:
        for j in range(k):
            ddts_[:, j] = basis.T @ f(basis @ states_[:, j])
    elif inputs.ndim == 1:
        for j in range(k):
            ddts_[:, j] = basis.T @ f(basis @ states_[:, j], inputs[j])
    else:
        for j in range(k):
            ddts_[:, j] = basis.T @ f(basis @ states_[:, j], inputs[:, j])

    return states_, ddts_



# models/_utils.py
"""Private utility functions for working with Model classes."""

__all__ = [
    "is_continuous",
    "is_discrete",
    "is_parametric",
    "is_interpolatory",
    "is_nonparametric",
]

from .mono._nonparametric import (
    ContinuousModel,
    DiscreteModel,
    _NonparametricModel,
)
from .mono._parametric import (
    _ParametricContinuousMixin,
    _ParametricDiscreteMixin,
    _ParametricModel,
    _InterpModel,
)


def is_continuous(model):
    """``True`` if the model is time continuous (semi-discrete)."""
    return isinstance(
        model,
        (ContinuousModel, _ParametricContinuousMixin),
    )


def is_discrete(model):
    """``True`` if the model is time discrete (fully discrete)."""
    return isinstance(
        model,
        (DiscreteModel, _ParametricDiscreteMixin),
    )


def is_nonparametric(model):
    """``True`` if the model is nonparametric."""
    return isinstance(model, _NonparametricModel)


def is_parametric(model):
    """``True`` if the model is parametric."""
    return isinstance(model, _ParametricModel)


def is_interpolatory(model):
    """``True`` if the model is fully interpolatory."""
    return isinstance(model, _InterpModel)



# lstsq/_base.py
"""Base class for solvers for the Operator Inference regression problem."""

__all__ = [
    "lstsq_size",
    "SolverTemplate",
    "PlainSolver",
]

import os
import abc
import copy
import warnings
import numpy as np
import scipy.linalg as la

from .. import errors, utils


_require_trained = utils.requires2(
    attr="data_matrix",
    message="solver not trained, call fit()",
)


def lstsq_size(modelform, r, m=0, affines=None) -> int:
    r"""Compute the number of columns in the operator matrix :math:`\Ohat` in
    the Operator Inference least-squares problem. This is also the number of
    columns in the data matrix :math:`\D`.

    Parameters
    ----------
    modelform : str containing 'c', 'A', 'H', 'G', and/or 'B'
        The structure of the desired reduced-order model. Each character
        indicates the presence of a different term in the model:
        'c' : Constant term c
        'A' : Linear state term Ax.
        'H' : Quadratic state term H(x⊗x).
        'G' : Cubic state term G(x⊗x⊗x).
        'B' : Input term Bu.
        For example, modelform=="AB" means f(x,u) = Ax + Bu.
    r : int
        The dimension of the reduced order model.
    m : int
        The dimension of the inputs of the model.
        Must be zero unless 'B' is in `modelform`.
    affines : dict(str -> list(callables))
        Functions that define the structures of the affine operators.
        Keys must match the modelform:
        * 'c': Constant term c(µ).
        * 'A': Linear state matrix A(µ).
        * 'H': Quadratic state matrix H(µ).
        * 'G': Cubic state matrix G(µ).
        * 'B': linear Input matrix B(µ).
        For example, if the constant term has the affine structure
        c(µ) = θ1(µ)c1 + θ2(µ)c2 + θ3(µ)c3, then 'c' -> [θ1, θ2, θ3].

    Returns
    -------
    ncols : int
        The number of columns in the Operator Inference least-squares problem.
    """
    if "B" in modelform and m == 0:
        raise ValueError("argument m > 0 required since 'B' in modelform")
    if "B" not in modelform and m != 0:
        raise ValueError(f"argument m={m} invalid since 'B' in modelform")

    if affines is None:
        affines = {}

    qs = [
        (
            len(affines[op])
            if (op in affines and op in modelform)
            else 1 if op in modelform else 0
        )
        for op in "cAHGB"
    ]
    rs = [1, r, r * (r + 1) // 2, r * (r + 1) * (r + 2) // 6, m]

    return sum(qq * rr for qq, rr in zip(qs, rs))


class SolverTemplate(abc.ABC):
    r"""Template for solvers for the Operator Inference regression
    :math:`\Z \approx \Ohat\D\trp` (or :math:`\D\Ohat\trp \approx \Z\trp`)
    for the operator matrix :math:`\Ohat`.

    Child classes formulate the regression, which may include regularization
    terms and/or optimization constraints. Hyperparameters should be set in
    the constructor (regularization terms, etc.).
    """

    def __init__(self):
        self.__D = None
        self.__Z = None

    # Properties: matrices ----------------------------------------------------
    @property
    def data_matrix(self) -> np.ndarray:
        r""":math:`k \times d` data matrix :math:`\D`."""
        return self.__D

    @property
    def lhs_matrix(self) -> np.ndarray:
        r""":math:`r \times k` left-hand side data :math:`\Z`."""
        return self.__Z

    # Properties: matrix dimensions -------------------------------------------
    @property
    def k(self) -> int:
        r"""Number of equations in the least-squares problem
        (number of rows of :math:`\D` and number of columns of :math:`\Z`).
        """
        return D.shape[0] if (D := self.data_matrix) is not None else None

    @property
    def d(self) -> int:
        r"""Number of unknowns in each row of the operator matrix
        (number of columns of :math:`\D` and :math:`\Ohat`).
        """
        return D.shape[1] if (D := self.data_matrix) is not None else None

    @property
    def r(self) -> int:
        r"""Number of operator matrix rows to learn
        (number of rows of :math:`\Z` and :math:`\Ohat`)
        """
        return Z.shape[0] if (Z := self.lhs_matrix) is not None else None

    # String representation ---------------------------------------------------
    def __str__(self) -> str:
        """String representation: class name + dimensions."""
        out = [self.__class__.__name__]
        if (self.data_matrix is not None) and (self.lhs_matrix is not None):
            out.append(f"  data_matrix:     {self.data_matrix.shape}")
            out.append(f"    condition number: {self.cond():.4e}")
            out.append(f"  lhs_matrix:      {self.lhs_matrix.shape}")
            out.append(f"  solve().shape:   {self.r, self.d}")
        else:
            out[0] += " (not trained)"
        return "\n".join(out)

    def __repr__(self) -> str:
        """Unique ID + string representation."""
        return utils.str2repr(self)

    @staticmethod
    def _print_kwargs(opts: dict) -> str:
        """Utility for printing routine options in __str__()."""
        return ", ".join([f"{key}={repr(val)}" for key, val in opts.items()])

    # Main methods -----------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        r"""Verify dimensions and save the data matrices.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
            If one-dimensional, assume :math:`r = 1`.
        """
        # Verify dimensions.
        if data_matrix.ndim != 2:
            raise ValueError("data_matrix must be two-dimensional")
        if lhs_matrix.ndim == 1:
            lhs_matrix = lhs_matrix.reshape((1, -1))
        if lhs_matrix.ndim != 2:
            raise ValueError("lhs_matrix must be one- or two-dimensional")
        if (k1 := lhs_matrix.shape[1]) != (k2 := data_matrix.shape[0]):
            raise errors.DimensionalityError(
                "data_matrix and lhs_matrix not aligned "
                f"(lhs_matrix.shape[-1] = {k1} != {k2} = data_matrix.shape[0])"
            )

        self.__D = data_matrix
        self.__Z = lhs_matrix

        return self

    @abc.abstractmethod
    def solve(self) -> np.ndarray:
        r"""Solve the Operator Inference regression.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        raise NotImplementedError  # pragma: no cover

    # Post-processing --------------------------------------------------------
    @_require_trained
    def cond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of the data matrix
        :math:`\D`.

        Returns
        -------
        conditionnumber : float
        """
        return np.linalg.cond(self.data_matrix)

    @_require_trained
    def residual(self, Ohat: np.ndarray) -> np.ndarray:
        r"""Compute the residual of the :math:`2`-norm regression objective for
        each row of the given operator matrix.

        Specifically, given a potential :math:`\Ohat`, compute

        .. math::
           \|\D\ohat_i - \z_i\|_2^2,
           \quad i = 1, \ldots, r,

        where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
        :math:`\Ohat` and :math:`\Z`, respectively.

        Parameters
        ----------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat`.

        Returns
        -------
        residuals : (r,) ndarray
            :math:`2`-norm residuals for each row of the operator matrix.
        """
        if self.r == 1 and Ohat.ndim == 1:
            Ohat = Ohat.reshape((1, -1))
        if Ohat.shape != (shape := (self.r, self.d)):
            raise errors.DimensionalityError(
                f"Ohat.shape = {Ohat.shape} != {shape} = (r, d)"
            )
        return np.sum(
            (self.data_matrix @ Ohat.T - self.lhs_matrix.T) ** 2,
            axis=0,
        )

    # Persistence -------------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):  # pragma: no cover
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        raise NotImplementedError

    @classmethod
    def load(cls, loadfile: str):  # pragma: no cover
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        solver : SolverTemplate
            Loaded solver.
        """
        raise NotImplementedError

    def _save_dict(self, hf, attr="options"):
        """Helper method for serializing a dictionary in HDF5 format."""
        options = hf.create_dataset(attr, shape=(0,))
        for key, value in getattr(self, attr).items():
            options.attrs[key] = "NULL" if value is None else value

    @staticmethod
    def _load_dict(hf, attr="options"):
        """Helper method for loading a dictionary from HDF5 format. This
        function is the inverse of :meth:`_save_dict()`."""
        options = dict()
        for key in hf[attr].attrs:
            value = hf[attr].attrs[key]
            options[key] = None if value == "NULL" else value
        return options

    def copy(self):
        """Make a copy of the solver."""
        return copy.deepcopy(self)

    # Verification ------------------------------------------------------------
    def verify(self):
        """Verify the solver.

        If the solver is already trained, check :meth:`solve()`,
        :meth:`copy()`, and (if implemented) :meth:`save()` and :meth:`load()`.
        If the solver is not trained, do nothing.
        """
        if self.data_matrix is None:
            return print("cannot verify solve() before calling fit()")

        tempfile = "_solververification.h5"

        def _make_copy(solver):
            newsolver = solver.copy()
            if (s2cls := newsolver.__class__) is not (scls := self.__class__):
                raise errors.VerificationError(
                    f"{scls.__name__}.copy() returned object "
                    f"of type '{s2cls.__name__}'"
                )
            return newsolver

        def _check_equality(obj1, obj2, Ohat1, operation):
            if obj1.r != obj2.r or obj1.d != obj2.d or obj1.k != obj2.k:
                raise errors.VerificationError(
                    f"{operation} does not preserve problem dimensions"
                )
            if not np.allclose(obj2.solve(), Ohat1):
                raise errors.VerificationError(
                    f"{operation} does not preserve the result of solve()"
                )

        try:
            # Verify solve().
            Ohat = self.solve()
            if (shape1 := Ohat.shape) != (shape2 := (self.r, self.d)):
                raise errors.VerificationError(
                    "solve() did not return array of shape (r, d) "
                    f"(expected {shape2}, got {shape1})"
                )

            # Verify copy().
            self2 = _make_copy(self)
            _check_equality(self, self2, Ohat, "copy()")
            print("solve() is consistent with problem dimensions")

            # Verify save()/load().
            try:
                self2.save(tempfile, overwrite=True)
                self3 = self2.load(tempfile)
                _check_equality(self2, self3, Ohat, "save()/load()")
            except NotImplementedError:
                print("save() and/or load() not implemented")
        finally:
            if os.path.isfile(tempfile):
                os.remove(tempfile)


class PlainSolver(SolverTemplate):
    r"""Solve the :math:`2`-norm ordinary least-squares problem without any
    regularization or constraints.

    That is, solve

    .. math::
        \argmin_{\Ohat} \|\D\Ohat\trp - \Z\trp\|_F^2.

    The solution is calculated using :func:`scipy.linalg.lstsq()`.

    Parameters
    ----------
    cond : float or None
        Cutoff for 'small' singular values of the data matrix.
        See :func:`scipy.linalg.lstsq()`.
    lapack_driver : str or None
        Which LAPACK driver is used to solve the least-squares problem.
        See :func:`scipy.linalg.lstsq()`.
    """

    def __init__(self, cond=None, lapack_driver=None):
        """Store least-squares solver options."""
        SolverTemplate.__init__(self)
        self.__options = dict(cond=cond, lapack_driver=lapack_driver)

    @property
    def options(self):
        """Keyword arguments for :func:`scipy.linalg.lstsq()`."""
        return self.__options

    def __str__(self):
        """String representation: dimensions + solver options."""
        start = SolverTemplate.__str__(self)
        kwargs = self._print_kwargs(self.options)
        return start + f"\n  solver: scipy.linalg.lstsq({kwargs})"

    # Main methods ------------------------------------------------------------
    def fit(self, data_matrix, lhs_matrix):
        r"""Verify dimensions and save the data matrices.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
            If one-dimensional, assume :math:`r = 1`.
        """
        super().fit(data_matrix, lhs_matrix)
        if self.k < self.d:
            warnings.warn(
                "least-squares system is underdetermined",
                errors.OpInfWarning,
            )
        return self

    def solve(self):
        r"""Solve the Operator Inference regression via
        :func:`scipy.linalg.lstsq()`.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        results = la.lstsq(self.data_matrix, self.lhs_matrix.T, **self.options)
        return results[0].T

    # Persistence -------------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            self._save_dict(hf, "options")
            if self.data_matrix is not None:
                hf.create_dataset("data_matrix", data=self.data_matrix)
                hf.create_dataset("lhs_matrix", data=self.lhs_matrix)

    @classmethod
    def load(cls, loadfile: str):
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        solver : SolverTemplate
            Loaded solver.
        """
        options = dict()
        with utils.hdf5_loadhandle(loadfile) as hf:

            options = cls._load_dict(hf, "options")
            solver = cls(**options)

            if "data_matrix" in hf:
                D = hf["data_matrix"][:]
                Z = hf["lhs_matrix"][:]
                solver.fit(D, Z)

        return solver

    def copy(self):
        """Make a copy of the solver."""
        solver = self.__class__(
            cond=self.options["cond"],
            lapack_driver=self.options["lapack_driver"],
        )
        if self.data_matrix is not None:
            SolverTemplate.fit(solver, self.data_matrix, self.lhs_matrix)
        return solver



# utils/__init__.py
r"""Miscellaneous utility functions."""


from ._gridsearch import *
from ._hdf5 import *
from ._mpl_config import *
from ._repr import *
from ._requires import *
from ._timer import *



# roms/_parametric.py
"""Parametric ROM classes."""

__all__ = [
    "ParametricROM",
]

from ..models import _utils as modutils
from ._base import _BaseROM


class ParametricROM(_BaseROM):
    r"""Parametric reduced-order model.

    This class connects classes from the various submodules to form a complete
    reduced-order modeling workflow.

    High-dimensional data
    :math:`\to` transformed / preprocessed data
    :math:`\to` compressed data
    :math:`\to` low-dimensional model.

    Parameters
    ----------
    model : :mod:`opinf.models` object
        Parametric system model, an instance of one of the following:

        * :class:`opinf.models.ParametricContinuousModel`
        * :class:`opinf.models.ParametricDiscreteModel`
        * :class:`opinf.models.InterpContinuousModel`
        * :class:`opinf.models.InterpDiscreteModel`
    lifter : :mod:`opinf.lift` object or None
        Lifting transformation.
    transformer : :mod:`opinf.pre` object or None
        Preprocesser.
    basis : :mod:`opinf.basis` object or None
        Dimensionality reducer.
    ddt_estimator : :mod:`opinf.ddt` object or None
        Time derivative estimator.
        Ignored if ``model`` is not time continuous.
    """

    def __init__(
        self,
        model,
        *,
        lifter=None,
        transformer=None,
        basis=None,
        ddt_estimator=None,
    ):
        """Store each argument as an attribute."""
        if not modutils.is_parametric(model):
            raise TypeError("'model' must be a parametric model instance")
        super().__init__(model, lifter, transformer, basis, ddt_estimator)

    # Training ----------------------------------------------------------------
    def fit(
        self,
        parameters,
        states,
        lhs=None,
        inputs=None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
    ):
        """Calibrate the model to training data.

        Parameters
        ----------
        parameters : list of s (floats or (p,) ndarrays)
            Parameter values for which training data are available.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is the data corresponding to parameter value
            ``parameters[i]``; each column ``states[i][:, j]`` is one snapshot.
        lhs : list of s (n, k_i) ndarrays or None
            Left-hand side regression data. Each array ``lhs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``lhs[i][:, j]`` corresponds to the snapshot ``states[i][:, j]``.

            - If the model is time continuous, these are the time derivatives
              of the state snapshots.
            - If the model is fully discrete, these are the "next states"
              corresponding to the state snapshots.

            If ``None``, these are estimated using :attr:`ddt_estimator`
            (time continuous) or extracted from ``states`` (fully discrete).
        inputs : list of s (m, k_i) ndarrays or None
            Input training data. Each array ``inputs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``inputs[i][:, j]`` corresponds to the snapshot ``states[:, j]``.
            May be a two-dimensional array if :math:`m=1` (scalar input).
            Only required if one or more model operators depend on inputs.
        fit_transformer : bool
            If ``True`` (default), calibrate the preprocessing transformation
            using the ``states``. If ``False``, assume the transformer is
            already calibrated.
        fit_basis : bool
            If ``True`` (default), calibrate the high-to-low dimensional
            mapping using the ``states``.
            If ``False``, assume the basis is already calibrated.

        Returns
        -------
        self
        """
        self._fit_and_return_training_data(
            parameters=parameters,
            states=states,
            lhs=lhs,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )
        return self

    # Evaluation --------------------------------------------------------------
    def predict(self, parameter, state0, *args, **kwargs):
        r"""Evaluate the reduced-order model.

        Arguments are the same as the ``predict()`` method of :attr:`model`.

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state0 : (n,) ndarray
            Initial state, expressed in the original state space.
        *args : list
            Other positional arguments to the ``predict()`` method of
            :attr:`model`.
        **kwargs : dict
            Keyword arguments to the ``predict()`` method of :attr:`model`.

        Returns
        -------
        states: (n, k) ndarray
            Solution to the model, expressed in the original state space.
        """
        q0_ = self.encode(state0, fit_transformer=False, fit_basis=False)
        states = self.model.predict(parameter, q0_, *args, **kwargs)
        return self.decode(states)
